% =========================================================================
% =====================  weetabixharry  22/03/2014  =======================
% =========================================================================
%
% Any polynomial of order N can be written in the form:
%
%               f(x) = a0 + a1*x + a2*x^2 + ... + aN*x^N                (1)
%
% Our numerical input data is assumed to be the first L samples of such a
% polynomial:
%
%                 y = [f(0), f(1), f(2), ... , f(L-1)]                  (2)
%
% and our task is to find the function f(x) which best describes this data.
%
% For the sake of simplicity, let us start by generating some dummy data
% from a known polynomial. That way, our solution will be easy to check.

close all; clear all; clc;

% Maximum order of polynomial
N = 3;

% Generate some input data
L = 24; % maximum x value (starting from 0)
x = (0:L).';
y = (1/3)*x.^3 - 10*x.^2 + 75*x - 33;
y = y + 10*randn(L+1,1); % add some random noise to make it less obvious

% In order to estimate the polynomial f(x) -- which, for this example, we
% know should be f(x) = 7*x.^3 - 43*x.^2 + 5 -- we can first look at the
% structure of the data values.
%
% By assumption (i.e. simply plugging x = 0, 1, 2... into (1)):
%
%       y[0] = a0
%       y[1] = a0 + a1 + ... + aN
%       y[2] = (2^0)*a0 + (2^1)*a1 + (2^2)*a2 + ... + (2^N)*aN
%       ...
%       y[n] = (n^0)*a0 + (n^1)*a1 + (n^2)*a2 + ... + (n^N)*aN
%       ...
%       y[L] = (L^0)*a0 + (L^1)*a1 + (L^2)*a2 + ... + (L^N)*aN          (3)
%
% We want to use the input data to solve for the coefficients a0, a1, ..., 
% aN (in a least squares sense). This will provide a "best fit" polynomial
% expression for the data (by substitution into (1)).
%
% To do this, we can arrange equation (3) into a standard "Ax = b" matrix
% equation. Or, using our notation:
%
%                              C*a = y                                  (4)
%
% where y is the input data (2) and "a" is the (N+1 x 1) column vector [a0,
% a1, ..., aN]. The (L x N+1) matrix, C, contains all the coefficients of
% a_n (for n = 0, 1, ..., N) for all of the rows of equation (3). A quick
% inspection reveals that the [n,p]th element of C must therefore be:
%
%                          C[n,p] = n^p                                 (5)
%
% for n = 0, 1, ..., L-1 and for p = 0, 1, ..., N. This is remarkably
% simple to construct using Matlab. We can either use "for" loops or,
% more efficiently, the "binary singleton expansion" function:

C = bsxfun(@power, (0:L).', (0:N));

% Finally, we just need to solve for "a" in equation (4). The well-known
% least-squares solution to this type of standard matrix equation is
% provided via the Moore-Penrose pseudoinverse:

a = pinv(C)*y;

% We can check our solution by plugging our values for a0, a1, ..., aN back
% into equation (4):

y_approx = C*a;

% and plotting this alongside the input data:

plot(x,y); grid on; hold on;
plot(x,y_approx,'--r');
xlabel('x'); ylabel('y');
legend('Input data', 'Polynomial Approximation');
